/*
 * RealTimeTaskImpl.hh
 *
 * Copyright (c) 2003, 2004 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file RealTimeTaskImpl.hh
 *
 * Header file for the basic implementation of Broker::RealTimeTask.
 */

#ifndef _real_time_task_impl_hh
#define _real_time_task_impl_hh

#include <rk.h>

#include "BrokerS.h"

/**
 * A base implementation of the Broker::RealTimeTask interface.  An instance of
 * this object should be the last link in an advocate chain since it will be
 * responsible for passing the report on to the manager through the
 * ChangeTaskCPU method.
 */
class RealTimeTaskImpl : public virtual POA_Broker::RealTimeTask
{
    
public:

    /**
     * Construct an RealTimeTaskImpl from the given parameters.  The current
     * set of recognized parameters are:
     *
     * @li @e name A string to be used as the name of the task.  (@e required)
     *
     * @param tp The list of parameters.
     *
     * @exception CORBA::SystemException for CORBA related errors.
     * @exception Broker::InvalidTaskParameter if one of the parameters is
     *            invalid.
     * @exception Broker::MissingTaskParameter if one of the required
     *            parameters is missing.
     */
    RealTimeTaskImpl(const Broker::TaskParameters &tp)
	throw (CORBA::SystemException,
	       Broker::DuplicateTaskParameter,
	       Broker::InvalidTaskParameter,
	       Broker::MissingTaskParameter);

    /**
     * Destruct an RealTimeTaskImpl.
     */
    virtual ~RealTimeTaskImpl();

    /** @copydoc Broker::Task::Name */
    virtual char *Name(void)
	throw (CORBA::SystemException);

    /** @copydoc Broker::Task::SetManager */
    virtual void SetManager(Broker::Manager_ptr man)
	throw (CORBA::SystemException);

    /** @copydoc Broker::Task::BeginCPUScheduling */
    virtual void BeginCPUScheduling(const Broker::ScheduleParameters &sp)
	throw (CORBA::SystemException,
	       Broker::DuplicateScheduleParameter,
	       Broker::InvalidScheduleParameter,
	       Broker::MissingScheduleParameter)
    {
    };
    
    /** @copydoc Broker::Task::EndCPUScheduling */
    virtual void EndCPUScheduling(void)
	throw (CORBA::SystemException)
    {
    };
    
    /** @copydoc Broker::RealTimeTask::PassCPU */
    virtual Broker::CPUReserve PassCPU(Broker::RealTimeTask_ptr rtt,
				       const Broker::CPUReserve &status,
				       const Broker::CPUReserve &advice,
				       const Broker::KeyedReportParameters &krp)
	throw (CORBA::SystemException);

    /** @copydoc Broker::RealTimeTask::ReportCPU */
    virtual Broker::CPUReserve ReportCPU(const Broker::CPUReserve &status,
					 const Broker::CPUReserve &advice,
					 const Broker::KeyedReportParameters &krp)
	throw (CORBA::SystemException)
    {
	return this->PassCPU(this->_this(), status, advice, krp);
    };

protected:

    /**
     * The name of the task and the resource set created for it.
     *
     * @invariant != NULL
     */
    CORBA::String_var rtti_Name;

    /**
     * The Broker::Manager used to handle scheduling during contention.  Only
     * valid between #BeginCPUScheduling and #EndCPUScheduling calls.
     */
    Broker::Manager_var rtti_Manager;
    
};

#endif
